from argparse import ArgumentParser
import importlib
import pkgutil
from pathlib import Path

from istari.constants import BASE_DIR
from istari.cli.base import BaseCommand
from istari.templates.plugins.base import BasePlugin


EXCLUDE_DIRS = ['.git', '__pycache__']

EXCLUDE_EXTS = ['.pyc', '.pyo', '.pyd', '.py.class', '.DS_Store']


class TemplateCommand(BaseCommand):
    template_name = None

    def get_template_name(self):
        assert self.template_name is not None, (
            f'{self.__class__.__name__} must define `.template_name` or implement `.get_template_name()`'
        )
        return self.template_name

    def get_target_path(self, template_path: Path) -> Path:
        target_path = str(self.target_dir / template_path.relative_to(self.template_dir))
        for var, val in self.variables.items():
            target_path = target_path.replace(f'{{{{ {var} }}}}', val)
        return Path(target_path)

    def get_path_mappings(self):
        paths = []
        for template_path in self.template_dir.rglob('*'):
            target_path = self.get_target_path(template_path)
            if template_path.is_dir() and target_path.name in self.exclude_dirs:
                continue
            if template_path.is_file() and target_path.name.endswith(tuple(self.exclude_exts)):
                continue
            paths.append((template_path, target_path))
        return paths

    def write_target_file(self, template_path, target_path):
        if target_path.suffix == '.py-tpl':
            target_path = target_path.with_suffix('.py')
        with open(template_path, 'r') as template_file:
            data = template_file.read()
        for var, val in self.variables.items():
            data = data.replace(f'{{{{ {var} }}}}', val)
        with open(target_path, 'w') as target_file:
            target_file.write(data)

    def process_template(self):
        paths = self.get_path_mappings()

        for template_path, target_path in paths:
            if target_path.exists():
                path_type = 'File' if template_path.is_file() else 'Directory'
                raise Exception(f'{path_type} {target_path.name} already exists.')
            
        for template_path, target_path in paths:
            if template_path.is_dir():
                target_path.mkdir()
            if template_path.is_file():
                self.write_target_file(template_path, target_path)

    def load_available_plugins(self, module_name: str) -> dict[str, BasePlugin]:
        plugins =  {
            plugin_name: self.load_plugin_class(f'{module_name}.{plugin_name}')
            for _, plugin_name, ispkg in pkgutil.iter_modules(importlib.import_module(module_name).__path__)
            if not ispkg and not plugin_name.startswith('_') and plugin_name != 'base'
        }
        return {k: v for k, v in plugins.items() if self.get_template_name().startswith(v.get_template())}

    def load_plugin_class(self, module_name: str) -> BasePlugin:
        module = importlib.import_module(module_name)
        plugin_class = getattr(module, 'Plugin')
        return plugin_class()

    def process_plugins(self, **options):
        for plugin, instance in self.plugins.items():
            flag = options.get(plugin, False)
            if flag is False:
                if input(f'{instance.get_help()}? (yes/no) [no]: ') in ['yes', 'y']:
                    flag = True
            if flag:
                instance.process(**options)

    def add_arguments(self, parser: ArgumentParser) -> None:
        self.plugins = self.load_available_plugins('istari.templates.plugins')
        for plugin, instance in self.plugins.items():
            flag = plugin.replace('_', '-')
            parser.add_argument(f'--{flag}', action='store_true', dest=plugin, help=instance.get_help())

    def handle(self, **options):
        options.setdefault('ignore_plugins', False)
        options.setdefault('target_dir', Path.cwd())

        self.template_dir = BASE_DIR / 'templates' / self.get_template_name()

        self.target_dir = options.get('target_dir')
        self.variables = options.get('variables', {})
        self.exclude_dirs = options.get('exclude_dirs', EXCLUDE_DIRS)
        self.exclude_exts = options.get('exclude_exts', EXCLUDE_EXTS)

        self.process_template()

        if options['ignore_plugins'] is False:
            self.process_plugins(**options)



import uuid
from django.core.exceptions import ValidationError
from django.db import models


class AutoCreatedField(models.DateTimeField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('auto_now_add', True)
        super().__init__(*args, **kwargs)


class AutoLastModifiedField(models.DateTimeField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('auto_now', True)
        super().__init__(*args, **kwargs)


class UUIDField(models.UUIDField):
    def __init__(self, primary_key=False, version=4, editable=False, *args, **kwargs):
        if version == 2:
            raise ValidationError('UUID version 2 is not supported.')
        
        if version < 1 or version > 5:
            raise ValidationError('UUID version is not valid.')
        
        if version == 1:
            default = uuid.uuid1
        elif version == 3:
            default = uuid.uuid3
        elif version == 4:
            default = uuid.uuid4
        elif version == 5:
            default == uuid.uuid5

        kwargs.setdefault('primary_key', primary_key)
        kwargs.setdefault('default', default)
        kwargs.setdefault('editable', editable)
        super().__init__(*args, **kwargs)



from argparse import ArgumentParser


class BaseCommand:
    def add_arguments(self, parser: ArgumentParser) -> None:
        return None
    
    def handle(self, **options):
        raise NotImplementedError(f'{self.__class__.__name__} must implement `.handle(self, **options)`')



class BasePlugin:
    help = None

    template = None

    def get_help(self):
        assert self.help is not None, (
            f'{self.__class__.__name__} must define `.help` or implement `.get_help()`'
        )
        return self.help

    def get_template(self):
        assert self.template is not None, (
            f'{self.__class__.__name__} must define `.template` or implement `.get_template()`'
        )
        return self.template

    def process(self, **options):
        raise NotImplementedError(f'{self.__class__.__name__} must implement `.process()`')



import secrets

def get_random_secret_key():
    '''
    Equivalent to django.core.management.utils.get_random_secret_key
    with the exception of excluding $ from the character set
    '''
    return ''.join(secrets.choice('abcdefghijklmnopqrstuvwxyz0123456789!@#%^&*(-_=+)') for i in range(50))



from pathlib import Path

from istari.constants import BASE_DIR
from istari.templates.plugins.base import BasePlugin


class Plugin(BasePlugin):
    help = 'Install istari in editable mode'

    template = 'project'

    def append_to_compose_volumes(self, value: str, path: Path) -> None:
        with open(path, 'r') as f:
            contents = f.readlines()    
        fp = 0
        for i, line in enumerate(contents):
            if line.lstrip().startswith('volumes'):
                fp = i + 1
                break
        pad = len(contents[fp]) - len(contents[fp].lstrip(' '))
        while contents[fp].lstrip().startswith('-'):
            fp += 1
        contents.insert(fp, f"{pad * ' '}- {value}\n")
        with open(path, 'w') as f:
            f.writelines(contents)

    def append_to_makefile_command(self, command: str, value: str, path: Path) -> None:
        with open(path, 'r') as f:
            contents = f.readlines()
        fp = 0
        for i, line in enumerate(contents):
            if line.startswith(f'.PHONY: {command}'):
                fp = i + 1
                break
        while not contents[fp].startswith('.PHONY'):
            fp += 1
        fp -= 1
        contents.insert(fp, f"\t{value}\n")
        with open(path, 'w') as f:
            f.writelines(contents)

    def process(self, **options):
        target_dir = options['target_dir']
        self.append_to_compose_volumes(
            f'{BASE_DIR.parent}:/usr/app/istari',
            target_dir / 'compose.yaml',
        )
        self.append_to_makefile_command(
            'up',
            'docker exec -it ${NAME}_django pip install -e /usr/app/istari',
            target_dir / 'Makefile'
        )



from django.apps import apps
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import AbstractUser, UserManager
from django.db import models
from django.utils.crypto import get_random_string
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from istari.db.models import UUIDMixin


class User(AbstractUser, UUIDMixin):
    class Meta:
        abstract = True


class EmailUserManager(UserManager):
    def _create_user(self, email: str, username: str | None, password: str | None, **extra_fields):
        '''
        Create and save user with the given email, username, and password.
        '''
        if not email:
            raise ValueError('The given email must be set.')
        email = self.normalize_email(email)
        if username is not None:
            # Load model from app registry versus get_user_model
            # so this manager method can be used in migrations
            GlobalUserModel = apps.get_model(self.model._meta.app_label, self.model._meta.object_name)
            username = GlobalUserModel.normalize_username(username)
        user = self.model(email=email, username=username, **extra_fields)
        user.password = make_password(password)
        user.save(using=self._db)
        return user
        
    def create_user(self, email: str, username: str | None=None, password: str | None=None, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, username, password, **extra_fields)

    def create_superuser(self, email: str, username: str | None=None, password: str | None=None, **extra_fields):
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')
        
        return self._create_user(email, username, password, **extra_fields)


class EmailUser(User):
    email = models.EmailField(_('email address'), unique=True)
    username = models.CharField(max_length=150, unique=True, blank=True)

    objects = EmailUserManager()

    EMAIL_FIELD = 'email'
    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    class Meta:
        abstract = True

    def __str__(self):
        return self.email
    
    def _generate_username(self):
        name = self.get_full_name()
        if len(name) == 0:
            name = self.email.split('@')[0]
        username = slugify(name)
        while self.__class__.objects.filter(username=username).exists():
            username = f'{username}-{get_random_string(length=4)}'
        return username
    
    def save(self, *args, **kwargs):
        if self.username is None:
            self.username = self._generate_username()
        super().save(*args, **kwargs)



from pathlib import Path

from istari.commands.startapp import Command as StartAppCommand
from istari.templates.plugins.base import BasePlugin


class Plugin(BasePlugin):
    help = 'Define custom AUTH_USER_MODEL'

    template = 'project'

    def create_users_app(self):
        StartAppCommand().handle(app_name='users', target_dir=self.target_dir, ignore_plugins=True)

    def add_users_model(self):
        path = self.target_dir / self.project_name / 'apps' / 'users' / 'models.py'
        contents = [
            'from django.db import models\n',
            '\n',
            'from istari.auth.models import EmailUser\n',
            '\n',
            '\n',
            'class User(EmailUser):\n',
            '    pass\n',
        ]
        with open(path, 'w') as f:
            f.writelines(contents)

    def define_auth_user_model(self):
        path = self.target_dir / self.project_name / 'settings.py'
        with open(path, 'r') as f:
            contents = f.readlines()
        fp = 0
        for i, line in enumerate(contents):
            if line.lstrip().startswith('ALLOWED_HOSTS'):
                fp = i + 1
                break
        contents.insert(fp, "\nAUTH_USER_MODEL = 'users.User'\n")
        with open(path, 'w') as f:
            f.writelines(contents)

    def process(self, **options):
        self.project_name: str = options['project_name']
        self.target_dir: Path = options['target_dir']
        self.create_users_app()
        self.add_users_model()
        self.define_auth_user_model()



from argparse import ArgumentParser

from istari import __version__
from istari.constants import SECRET_KEY_INSECURE_PREFIX
from istari.templates.commands import TemplateCommand
from istari.utils import get_random_secret_key


class Command(TemplateCommand):
    template_name = 'project_template'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument('project_name')
        super().add_arguments(parser)
    
    def handle(self, **options):
        options['variables'] = {
            'docs_version': '5.0',
            'istari_version': __version__,
            'project_name': options['project_name'],
            'secret_key': SECRET_KEY_INSECURE_PREFIX + get_random_secret_key(),
        }
        return super().handle(**options)



from argparse import ArgumentParser
from pathlib import Path

from istari.templates.commands import TemplateCommand


class Command(TemplateCommand):
    template_name = 'app_template'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument('app_name')
        super().add_arguments(parser)

    def sort_installed_apps(self, apps: list[str]) -> list[str]:
        '''
        Sort INSTALLED_APPS by alphabetically
        sorting any apps that start with 'apps'
        '''
        start = [i for i, a in enumerate(apps) if a.startswith("'apps")][0]
        return apps[:start] + sorted(apps[start:])

    def append_installed_apps(self, settings: Path, app_name: str):
        with open(settings, 'r') as f:
            contents = f.readlines()

        # append app to INSTALLED_APPS
        start = 0
        for i, line in enumerate(contents):
            if line.startswith('INSTALLED_APPS'):
                start = i + 1
                break
        end = start
        while not contents[end].startswith(']'):
            end += 1
        contents.insert(end, f"    'apps.{app_name}',\n")
        end += 1

        # sort alphabetically
        apps = self.sort_installed_apps([app.strip() for app in contents[start:end]])
        del contents[start:end]
        for app in reversed(apps):
            contents.insert(start, f'    {app}\n')

        with open(settings, 'w') as f:
            f.writelines(contents)

    def handle(self, **options):
        options['target_dir'] = options.get('target_dir', Path.cwd())
        options['variables'] = {
            'app_name': options['app_name'],
        }

        if not (options['target_dir'] / 'manage.py').is_file():
            raise Exception('File manage.py not found. Please run in root directory of Django project.')
        
        settings = None
        for path in options['target_dir'].rglob('*'):
            if path.is_dir() and (path / 'settings.py').is_file():
                settings = path / 'settings.py'
                options['target_dir'] = path / 'apps'
                options['target_dir'].mkdir(exist_ok=True)
                break

        self.append_installed_apps(settings, options['app_name'])
        return super().handle(**options)



from setuptools import setup, find_packages


def get_version():
    for line in open('istari/__init__.py', 'r'):
        if line.startswith('__version__'):
            return line.split("'")[1]


setup(
    name='istari',
    version=get_version(),
    author='Janus Digital LLC',
    author_email='justice@janusdigital.io',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'istari=istari.cli:main',
        ],
    },
)



import importlib
import pkgutil
import sys
from argparse import ArgumentParser
from gettext import gettext as _

from istari import __version__
from istari.cli.base import BaseCommand


class CustomParser(ArgumentParser):
    def error(self, message):
        if message.startswith('argument command_name: invalid choice'):
            return

        if message.startswith('unrecognized arguments'):
            return

        self.print_usage(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(2, _('%(prog)s: error: %(message)s\n') % args)


class IstariCLI:
    def __init__(self):
        self.parser = CustomParser()
        self.parser.add_argument('-v', '--version', action='store_true')
        self.subparsers = self.parser.add_subparsers(title='commands', dest='command_name')

        self.command_map: dict[str, BaseCommand] = {}
        self.process_commands()

    def run(self):
        args = self.parser.parse_args().__dict__

        if args.pop('version', False):
            print(__version__)
            return

        command_name = args.pop('command_name', None)
        if command_name is None:
            print('\nAvailable commands:\n')
            for command in self.command_map.keys():
                print(command)
            print('')
            return

        self.command_map[command_name].handle(**args)

    def process_commands(self):
        self.command_map = self.get_commands('istari.commands')
        for name, instance in self.command_map.items():
            command_parser = self.subparsers.add_parser(name)
            instance.add_arguments(command_parser)

    def get_commands(self, module_name: str) -> dict[str, BaseCommand]:
        return {
            command_name: self.load_command_class(f'{module_name}.{command_name}')
            for _, command_name, ispkg in pkgutil.iter_modules(importlib.import_module(module_name).__path__)
            if not ispkg and not command_name.startswith('_')
        }
    
    def load_command_class(self, module_name: str) -> BaseCommand:
        module = importlib.import_module(module_name)
        command_class = getattr(module, 'Command')
        return command_class()


def main():
    IstariCLI().run()



from django.db import models
from django.template.defaultfilters import slugify
from django.utils.translation import gettext_lazy as _

from istari.db.fields import AutoCreatedField, AutoLastModifiedField, UUIDField


class SlugMixin(models.Model):
    slug = models.SlugField()

    SLUG_FIELD = 'name'

    def slugify(self):
        return slugify(getattr(self, self.SLUG_FIELD))
    
    def save(self, *args, **kwargs):
        if self.slug is None:
            self.slug = self.slugify()
        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class TimeStampedMixin(models.Model):
    created_at = AutoCreatedField()
    updated_at = AutoLastModifiedField()

    def save(self, *args, **kwargs):
        update_fields = kwargs.get('update_fields', None)
        if update_fields is not None:
            kwargs['update_fields'] = set(update_fields).union('updated_at')
        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class UUIDMixin(models.Model):
    uuid = UUIDField(_('UUID'))

    class Meta:
        abstract = True


class SlugModel(SlugMixin, TimeStampedMixin):
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True


class UUIDModel(TimeStampedMixin, UUIDMixin):
    class Meta:
        abstract = True
